@@ -15,6 +15,7 @@
 #endif
 #include <assert.h>
 #include <errno.h>
+#include <setjmp.h>
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -107,6 +108,10 @@ static int dump_triplets;
 /* Some Berkeley DB errors are not really error.  Keep them quiet. */
 static int muffle_error = 0;
 
+/* When Greyfix detects a problem, let emails pass through and log. */
+static jmp_buf defect_jmp_buf;
+static const char *defect_msg = 0;
+
 /**********************************************************************
  * Berkeley DB routines
  */
@@ -158,8 +163,11 @@ prepare_env()
 	dbenv->set_errcall(dbenv, db_errcall_fcn);
 	rc = dbenv->open(dbenv, db_home,
 			 DB_INIT_LOCK | DB_INIT_MPOOL | DB_CREATE, 0);
-	if (rc)
+	if (rc) {
 	    log_db_error("dbenv->open", rc);
+	    dbenv->close(dbenv, 0);
+	    dbenv = 0;
+	}
     }
     return rc;
 }
@@ -173,8 +181,11 @@ prepare_db()
 	log_db_error("db_create", rc);
     else {
 	rc = db_open(db, DB_FILE_NAME, NULL, DB_BTREE, DB_CREATE, 0644);
-	if (rc)
+	if (rc) {
 	    log_db_error("db->open", rc);
+	    db->close(db, 0);
+	    db = 0;
+	}
     }
     return rc;
 }
@@ -255,9 +266,7 @@ cleanup()
 static void
 fatal(const char *msg)
 {
-    int err = errno;
-    cleanup();
-    syslog(LOG_ERR, "fatal: %s: %s", msg, db_strerror(err));
+    syslog(LOG_ERR, "fatal: %s", msg);
     abort();
 }
 
@@ -268,6 +277,8 @@ run_expiry()
     int rc;
     time_t now;
     unsigned int count = 0;
+    if (db == 0)
+	return;
     if (time(&now) == (time_t)-1) {
 	syslog(LOG_ERR, "time failed during run_expiry");
 	return;
@@ -433,17 +444,25 @@ build_triplet_key(const char *ip, const char *from, const char *to)
     assert(dbkey.size <= dbkey.ulen);
 }
 
+/* If there is an error while processing a SMTP request, log a warning
+   in Postfix. */
+static void
+jmperr(const char *msg)
+{
+    defect_msg = msg;
+    longjmp(defect_jmp_buf, 1);
+}
+
 static void
 touch_data()
 {
     if (time(&triplet_data.access_time) == (time_t)-1)
-	fatal("time failed");
+	jmperr("time failed");
 }
 
 static void
 build_data()
 {
-    touch_data();
     triplet_data.create_time = triplet_data.access_time;
     triplet_data.block_count = 0;
     triplet_data.pass_count = 0;
@@ -453,16 +472,20 @@ static void
 get_grey_data()
 {
     int rc;
+    if (!db)
+	jmperr(PACKAGE_STRING " database not opened");
     rc = db->get(db, NULL, &dbkey, &dbdata, 0);
-    if (rc == DB_NOTFOUND)
+    if (rc == DB_NOTFOUND) {
+	touch_data();
 	build_data();
+    }
     else if (rc) {
 	log_db_error("get failed", rc);
-	fatal("Exiting");
+	jmperr("get failed");
     }
     else {
-	time_t ref_time, now;
-	double age_max, age;
+	time_t ref_time;
+	double age_max;
 	if (triplet_data.pass_count) {
 	    ref_time = triplet_data.access_time;
 	    age_max = pass_max_idle;
@@ -471,11 +494,10 @@ get_grey_data()
 	    ref_time = triplet_data.create_time;
 	    age_max = bloc_max_idle;
 	}
-	age = difftime(now, ref_time);
-	if (age > age_max)
+	touch_data();
+	/* Expire IDLE records */
+	if (difftime(triplet_data.access_time, ref_time) > age_max)
 	    build_data();
-	else
-	    touch_data();
     }
 }
 
@@ -532,6 +554,15 @@ static void
 process_smtp_rcpt()
 {
     double delay;
+    if (setjmp(defect_jmp_buf)) {
+	if (defect_msg) {
+	    printf(STR_ACTION "WARN %s\n", defect_msg);
+	    defect_msg = 0;
+	}
+	else
+	    puts(STR_ACTION "WARN " PACKAGE_STRING " is not working properly");
+	return;
+    }
     get_grey_data();
     delay = difftime(triplet_data.access_time, triplet_data.create_time);
     /* Block inbound mail that is from a previously unknown (ip, from, to) triplet */
@@ -778,8 +809,8 @@ main(int argc, const char **argv)
 
     rc = initialize();
     if (rc) {
-	errno = rc;
-	fatal("initialization failure");
+	cleanup();
+	syslog(LOG_ERR, "error: initialization failure: %s", db_strerror(rc));
     }
     if (dump_triplets)
 	do_dump_triplets();
